--- /dev/null	Sun Aug 11 00:54:16 2013
+++ new/usr/src/uts/common/sys/fsh_impl.h	Sun Aug 11 00:54:16 2013
@@ -0,0 +1,121 @@
+/*
+ * This file and its contents are supplied under the terms of the
+ * Common Development and Distribution License ("CDDL"), version 1.0.
+ * You may only use this file in accordance with the terms of version
+ * 1.0 of the CDDL.
+ *
+ * A full copy of the text of the CDDL should have accompanied this
+ * source.  A copy of the CDDL is also available via the Internet at
+ * http://www.illumos.org/license/CDDL.
+ */
+
+/*
+ * Copyright 2013 Damian Bogel.  All rights reserved.
+ */
+
+/*
+ * This file includes all the necessary declarations for fsh.c, vfs.c
+ * and vnode.c.
+ */
+
+#ifndef _FSH_IMPL_H
+#define	_FSH_IMPL_H
+
+#include <sys/atomic.h>
+#include <sys/fsh.h>
+#include <sys/list.h>
+#include <sys/pathname.h>
+#include <sys/types.h>
+#include <sys/vfs.h>
+#include <sys/vnode.h>
+
+#ifdef __cplusplus
+extern "C" {
+#endif
+
+#define	FSH_VFS_MOUNT		0
+#define	FSH_VFS_UNMOUNT		1
+
+#define	FSH_VOP_READ		2
+#define	FSH_VOP_WRITE		3
+
+#define	FSH_SUPPORTED_OPS_COUNT	4
+
+/*
+ * TODO:
+ * - change all defined lists to the ones used in illumos kernel
+ * - add comments
+ */
+
+typedef union fsh_fn {
+	FSH_OPS;
+} fsh_fn_t;
+
+/*
+ * fsh_int_t is the internal fsh_t structure. Instead of using a big
+ * structure containing hooks for all the operations, fsh_int_t contains
+ * a pointer (in union fsh_fn_t) to one function. Thanks to this approach,
+ * we don't have to manipulate such big structures. Instead, we keep
+ * a fsh_int_t list for every operation.
+ */
+struct fsh_int {
+	fsh_fn_t	fshi_fn;
+	void		*fshi_arg;
+	list_node_t	fshi_next;
+};
+/* typedef struct fsh_int fsh_int_t; in fsh.h */
+
+/*
+ * fshl_lock is read-locked by every call to a fsh_vop/vfsop() for
+ * entire execution. This way, we guarantee that a list of hooks is unchanged
+ * during one fop_foo()/fsop_foo() execution.
+ */
+typedef struct fsh_list {
+	krwlock_t	fshl_lock;
+	list_t		fshl_list;
+} fsh_list_t;
+
+/*
+ * fshfsr_enabled informs whether FSH is enabled for this vfs_t.
+ * There are no locks involved. This field should be change only via
+ * atomic operations.
+ */
+typedef struct fsh_fsrecord {
+	volatile uint_t	fshfsr_enabled;
+	fsh_list_t	fshfsr_opv[FSH_SUPPORTED_OPS_COUNT];
+} fsh_fsrecord_t;
+
+typedef struct fsh_callback_int {
+	fsh_callback_t	fshci_callback;
+	list_node_t	fshci_next;
+} fsh_callback_int_t;
+
+typedef struct fsh_callback_list {
+	krwlock_t	fshcl_lock;
+	list_t		fshcl_list;
+} fsh_callback_list_t;
+
+/* API for vnode.c and vfs.c only */
+/* vnode.c */
+extern int fsh_read(vnode_t *vp, uio_t *uiop, int ioflag, cred_t *cr,
+		caller_context_t *ct);
+extern int fsh_write(vnode_t *vp, uio_t *uiop, int ioflag, cred_t *cr,
+		caller_context_t *ct);
+
+/* vfs.c */
+extern void fsh_init(void);
+
+extern int fsh_mount(vfs_t *vfsp, vnode_t *mvp, struct mounta *uap,
+		cred_t *cr);
+extern int fsh_unmount(vfs_t *vfsp, int flag, cred_t *cr);
+
+extern void fsh_exec_mount_callbacks(vfs_t *vfsp);
+extern void fsh_exec_free_callbacks(vfs_t *vfsp);
+
+extern void fsh_fsrec_destroy(struct fsh_fsrecord *volatile fsrecp);
+
+#ifdef __cplusplus
+}
+#endif
+
+#endif /* _FSH_IMPL_H */
